import os
import logging
import sys
import json
import importlib
import importlib.util
import inspect

from typing import Tuple, Callable, Union, List, Dict, Optional
from abc import ABC, abstractmethod
from colorama import Fore

from label_studio_tools.core.label_config import parse_config
from label_studio_tools.core.utils.io import get_local_path
from .cache import create_cache

logger = logging.getLogger(__name__)

CACHE = create_cache(
    os.getenv('CACHE_TYPE', 'sqlite'),
    path=os.getenv('MODEL_DIR', '.'))

# Decorator to register predict function
_predict_fn: Callable
_update_fn: Callable


def predict_fn(f):
    global _predict_fn
    _predict_fn = f
    logger.info(f'{Fore.GREEN}Predict function "{_predict_fn.__name__}" registered{Fore.RESET}')
    return f


def update_fn(f):
    global _update_fn
    _update_fn = f
    logger.info(f'{Fore.GREEN}Update function "{_update_fn.__name__}" registered{Fore.RESET}')
    return f


class LabelStudioMLBase(ABC):
    TRAIN_EVENTS = (
        'ANNOTATION_CREATED',
        'ANNOTATION_UPDATED',
        'ANNOTATION_DELETED',
        'PROJECT_UPDATED'
    )

    def __init__(self, **kwargs) -> None:
        self.project_id = kwargs.get("project_id") or ''
        self._label_config = kwargs.get("label_config") or None

        if self._label_config is not None:
            print(f"Initializing Label config: \r {self._label_config}")
            print(
                "Initializing Parse Label Config: \r "
                f"{json.dumps(parse_config(self._label_config), indent=4, ensure_ascii=False)}")
            self.set('label_config', self._label_config)
            self.set('parsed_label_config', json.dumps(parse_config(self._label_config)))

    def use_label_config(self, label_config: str):
        current_label_config = self.get('label_config')
        if not current_label_config:
            # first time model is initialized
            self.set('model_version', 'INITIAL')
        if current_label_config != label_config:
            # label config has been changed
            self.set('label_config', label_config)
            self.set('parsed_label_config', json.dumps(parse_config(label_config)))

    def get(self, key: str):
        return CACHE[self.project_id, key]

    def set(self, key: str, value: str):
        CACHE[self.project_id, key] = value

    def has(self, key: str):
        return (self.project_id, key) in CACHE

    @property
    def label_config(self):
        return self.get('label_config')

    @property
    def parsed_label_config(self):
        return json.loads(self.get('parsed_label_config'))

    @property
    def model_version(self):
        return self.get('model_version')

    # @abstractmethod
    def predict(self, tasks: List[Dict], context: Optional[Dict] = None, **kwargs) -> List[Dict]:
        """
        Predict method should return a list of dicts with predictions for each task.
        :param tasks: a list of tasks
        :param context: a dict with additional context for model
        :param kwargs:
        :return: the list of dicts with predictions
        """

        # if there is a registered predict function, use it
        if _predict_fn:
            return _predict_fn(tasks, context, helper=self, **kwargs)

    def process_event(self, event, data, job_id, additional_params):
        if event in self.TRAIN_EVENTS:
            logger.debug(f'Job {job_id}: Received event={event}: calling {self.__class__.__name__}.fit()')
            train_output = self.fit(event=event, data=data, job_id=job_id, **additional_params)
            logger.debug(f'Job {job_id}: Train finished.')
            return train_output

    def fit(self, event, data, **additional_params):
        # if there is a registered update function, use it
        if _update_fn:
            return _update_fn(event, data, helper=self, **additional_params)

    def get_local_path(self, url, project_dir=None):
        return get_local_path(url, project_dir=project_dir, hostname=self.hostname, access_token=self.access_token)

    def get_first_tag_occurence(
            self,
            control_type: Union[str, Tuple],
            object_type: Union[str, Tuple],
            name_filter: Optional[Callable] = None,
            to_name_filter: Optional[Callable] = None
    ) -> Tuple[str, str, str]:
        """
        Reads config and returns the first control tag and the first object tag that match the given types
        For example: get_first_tag_occurence('Choices', 'Text') will return the first Choices tag that has an Text tag as input
        :param control_type:
        :param object_type:
        :param name_filter: if given, only tags with this name will be considered, e.g. name_filter=lambda name: name.startswith('my_')
        :param to_name_filter: if given, only tags with this name will be considered, e.g. to_name_filter=lambda name: toName.startswith('my_')
        :return: tuple of (from_name, to_name, value)
        """
        parsed_label_config = self.parsed_label_config
        for from_name, info in parsed_label_config.items():
            control_type_mathes = isinstance(control_type, str) and info['type'] == control_type or \
                                  isinstance(control_type, tuple) and info['type'] in control_type
            control_name_matches = name_filter is None or name_filter(from_name)

            if control_type_mathes and control_name_matches:

                for input_name, input in zip(info['to_name'], info['inputs']):
                    object_type_matches = isinstance(object_type, str) and input['type'] == object_type or \
                                          isinstance(object_type, tuple) and input['type'] in object_type
                    object_name_matches = to_name_filter is None or to_name_filter(input_name)

                    if object_type_matches and object_name_matches:
                        return from_name, info['to_name'][0], input['value']
        raise ValueError(
            f'No control tag of type {control_type} and object tag of type {object_type} found in label config')


def get_all_classes_inherited_LabelStudioMLBase(script_file):
    names = set()
    abs_path = os.path.abspath(script_file)
    module_name = os.path.splitext(os.path.basename(script_file))[0]
    sys.path.append(os.path.dirname(abs_path))
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        print(Fore.RED + 'Can\'t import module "' + module_name + f'", reason: {e}.\n'
                                                                  'If you are looking for examples, you can find a dummy model.py here:\n' +
              Fore.LIGHTYELLOW_EX + 'https://labelstud.io/tutorials/dummy_model.html')
        module = None
        exit(-1)

    for name, obj in inspect.getmembers(module, inspect.isclass):
        if name == LabelStudioMLBase.__name__:
            continue
        if issubclass(obj, LabelStudioMLBase):
            names.add(name)
        for base in obj.__bases__:
            if LabelStudioMLBase.__name__ == base.__name__:
                names.add(name)
    sys.path.pop()
    return list(names)



import os
import sqlite3
from abc import ABC, abstractmethod
from threading import Lock
from functools import lru_cache


class BaseCache(ABC):

    def __init__(self, path):
        self.path = path

    @abstractmethod
    def __getitem__(self, project_id_key: tuple):
        """
        Get value from cache
        :param project_id_key: tuple (project_id, key)
        :return:
        """

    @abstractmethod
    def __setitem__(self, project_id_key: tuple, value):
        """
        Set value to cache
        :param project_id_key: tuple (project_id, key)
        :param value:
        :return:
        """

    @abstractmethod
    def __contains__(self, project_id_key: tuple):
        """
        Check if value exists in cache
        :param project_id_key: tuple (project_id, key)
        :return:
        """

    @abstractmethod
    def __delitem__(self, project_id_key: tuple):
        """
        Delete value from cache
        :param project_id_key: tuple (project_id, key)
        :return:
        """


class SqliteCache(BaseCache):
    def __init__(self, path: str, db_name: str = 'cache.db'):
        super(SqliteCache, self).__init__(path)
        os.makedirs(self.path, exist_ok=True)
        self.db_name = os.path.join(self.path, db_name)
        self.lock = Lock()

        # Establish a connection and create table if it doesn't exist
        with self.lock, sqlite3.connect(self.db_name) as conn:
            cursor = conn.cursor()
            cursor.execute('''
                CREATE TABLE IF NOT EXISTS cache (
                    project_id TEXT NOT NULL,
                    key TEXT NOT NULL,  
                    value TEXT NOT NULL,
                    PRIMARY KEY (project_id, key)
                );
            ''')

    @lru_cache(maxsize=100)
    def __getitem__(self, project_id_key):
        project_id, key = project_id_key
        with self.lock, sqlite3.connect(self.db_name) as conn:
            cursor = conn.cursor()
            cursor.execute(
                'SELECT value FROM cache WHERE project_id = ? AND key = ?;',
                (project_id, key))
            result = cursor.fetchone()
            if result is None:
                return result
            return result[0]

    def __setitem__(self, project_id_key, value):
        project_id, key = project_id_key
        if not isinstance(value, str):
            raise ValueError('Value must be a string')
        with self.lock, sqlite3.connect(self.db_name) as conn:
            cursor = conn.cursor()
            cursor.execute('REPLACE INTO cache (project_id, key, value) VALUES (?, ?, ?);',
                           (project_id, key, value))
        self.__getitem__.cache_clear()

    def __delitem__(self, project_id_key):
        project_id, key = project_id_key
        with self.lock, sqlite3.connect(self.db_name) as conn:
            cursor = conn.cursor()
            cursor.execute('DELETE FROM cache WHERE project_id = ? AND key = ?;',
                           (project_id, key))
        self.__getitem__.cache_clear()

    def __contains__(self, project_id_key):
        project_id, key = project_id_key
        with self.lock, sqlite3.connect(self.db_name) as conn:
            cursor = conn.cursor()
            cursor.execute('SELECT 1 FROM cache WHERE project_id = ? AND key = ?;',
                           (project_id, key))
            return cursor.fetchone() is not None


def create_cache(cache_type, path, **kwargs):
    if cache_type == 'sqlite':
        return SqliteCache(path, **kwargs)
    else:
        raise ValueError(f"Unsupported cache type: {cache_type}")



import os
import argparse
import json
import logging
import logging.config

from flask_ngrok import run_with_ngrok

logging.config.dictConfig({
  "version": 1,
  "formatters": {
    "standard": {
      "format": "[%(asctime)s] [%(levelname)s] [%(name)s::%(funcName)s::%(lineno)d] %(message)s"
    }
  },
  "handlers": {
    "console": {
      "class": "logging.StreamHandler",
      "level": os.getenv('LOG_LEVEL'),
      "stream": "ext://sys.stdout",
      "formatter": "standard"
    }
  },
  "root": {
    "level": os.getenv('LOG_LEVEL'),
    "handlers": [
      "console"
    ],
    "propagate": True
  }
})

from label_studio_ml.api import init_app
from .model import LabelStudioMLBase

app = None


def run_ml_backend(port=9090, host='0.0.0.0', debug=False):
    global app
    app = init_app(model_class=LabelStudioMLBase)
    run_with_ngrok(app)
    app.run(host=host, port=port, debug=debug)



import json
import logging

from flask import Flask, request, jsonify

from .model import LabelStudioMLBase
from .exceptions import exception_handler


logger = logging.getLogger(__name__)

_server = Flask(__name__)
MODEL_CLASS = LabelStudioMLBase


def init_app(model_class):
    global MODEL_CLASS

    if not issubclass(model_class, LabelStudioMLBase):
        raise ValueError('Inference class should be the subclass of ' + LabelStudioMLBase.__class__.__name__)

    MODEL_CLASS = model_class
    return _server


@_server.route('/predict', methods=['POST'])
@exception_handler
def _predict():
    """
    Predict tasks

    Example request:
    request = {
            'tasks': tasks,
            'model_version': model_version,
            'project': '{project.id}.{int(project.created_at.timestamp())}',
            'label_config': project.label_config,
            'params': {
                'login': project.task_data_login,
                'password': project.task_data_password,
                'context': context,
            },
        }

    @return:
    Predictions in LS format
    """
    data = request.json
    print(f'Request: {json.dumps(data,ensure_ascii=False,indent=4)}')
    tasks = data.get('tasks')
    params = data.get('params') or {}
    project = data.get('projectUuid')
    if project:
        project_id = data.get('projectUuid')
    else:
        project_id = None
    label_config = data.get('label_config')
    context = params.pop('context', {})

    model = MODEL_CLASS(project_id=project_id)
    model.use_label_config(label_config)

    predictions = model.predict(tasks, context=context, **params)
    print(f'Predictions: {json.dumps(predictions,ensure_ascii=False,indent=4)}')
    return jsonify({'results': predictions})


@_server.route('/setup', methods=['POST'])
@exception_handler
def _setup():
    data = request.json
    print(f'data: {json.dumps(data, indent=4)}')
    project_id = data.get('projectUuid')
    print(f'project_id: {project_id}')
    label_config = data.get('schema')
    model = MODEL_CLASS(project_id=project_id, label_config=label_config)
    model.use_label_config(label_config)
    model_version = model.get('model_version')
    return jsonify({'model_version': model_version})


TRAIN_EVENTS = (
    'ANNOTATION_CREATED',
    'ANNOTATION_UPDATED',
    'ANNOTATION_DELETED',
    'PROJECT_UPDATED'
)


@_server.route('/webhook', methods=['POST'])
def webhook():
    data = request.json
    event = data.pop('action')
    if event not in TRAIN_EVENTS:
        return jsonify({'status': 'Unknown event'}), 200
    project_id = str(data['project']['id'])
    label_config = data['project']['label_config']
    model = MODEL_CLASS(project_id)
    model.use_label_config(label_config)
    model.fit(event, data)
    return jsonify({}), 201


@_server.route('/health', methods=['GET'])
@_server.route('/', methods=['GET'])
@exception_handler
def health():
    return jsonify({
        'status': 'UP',
        'model_class': MODEL_CLASS.__name__
    })


@_server.route('/metrics', methods=['GET'])
@exception_handler
def metrics():
    return jsonify({})


@_server.errorhandler(FileNotFoundError)
def file_not_found_error_handler(error):
    logger.warning('Got error: ' + str(error))
    return str(error), 404


@_server.errorhandler(AssertionError)
def assertion_error(error):
    logger.error(str(error), exc_info=True)
    return str(error), 500


@_server.errorhandler(IndexError)
def index_error(error):
    logger.error(str(error), exc_info=True)
    return str(error), 500


@_server.before_request
def log_request_info():
    logger.debug('Request headers: %s', request.headers)
    logger.debug('Request body: %s', request.get_data())


@_server.after_request
def log_response_info(response):
    logger.debug('Response status: %s', response.status)
    logger.debug('Response headers: %s', response.headers)
    logger.debug('Response body: %s', response.get_data())
    return response



import logging
import difflib

from PIL import Image, ImageOps
from collections import OrderedDict
from typing import List

from label_studio_tools.core.utils.params import get_env
from label_studio_tools.core.utils.io import get_local_path

DATA_UNDEFINED_NAME = '$undefined$'

logger = logging.getLogger(__name__)


def get_single_tag_keys(parsed_label_config, control_type, object_type):
    """
    Gets parsed label config, and returns data keys related to the single control tag and the single object tag schema
    (e.g. one "Choices" with one "Text")
    :param parsed_label_config: parsed label config returned by "label_studio.misc.parse_config" function
    :param control_type: control tag str as it written in label config (e.g. 'Choices')
    :param object_type: object tag str as it written in label config (e.g. 'Text')
    :return: 3 string keys and 1 array of string labels: (from_name, to_name, value, labels)
    """
    assert len(parsed_label_config) == 1
    from_name, info = list(parsed_label_config.items())[0]
    assert info['type'] == control_type, 'Label config has control tag "<' + info['type'] + '>" but "<' + control_type + '>" is expected for this model.'  # noqa

    assert len(info['to_name']) == 1
    assert len(info['inputs']) == 1
    assert info['inputs'][0]['type'] == object_type
    to_name = info['to_name'][0]
    value = info['inputs'][0]['value']
    return from_name, to_name, value, info['labels']


def get_first_tag_keys(parsed_label_config, control_type, object_type):
    """
    Reads config and returns the first control tag and the first object tag that match the given types
    :param parsed_label_config:
    :param control_type:
    :param object_type:
    :return:
    """
    for from_name, info in parsed_label_config.items():
        if info['type'] == control_type:
            for input in info['inputs']:
                if input['type'] == object_type:
                    return from_name, info
    return None, None


def is_skipped(completion):
    if len(completion['annotations']) != 1:
        return False
    completion = completion['annotations'][0]
    return completion.get('skipped', False) or completion.get('was_cancelled', False)


def get_choice(completion):
    return completion['annotations'][0]['result'][0]['value']['choices'][0]


def get_image_local_path(url, image_cache_dir=None, project_dir=None, image_dir=None,
                         label_studio_host=None, label_studio_access_token=None):
    image_local_path = get_local_path(
        url=url,
        cache_dir=image_cache_dir,
        project_dir=project_dir,
        hostname=label_studio_host or get_env('HOSTNAME'),
        image_dir=image_dir,
        access_token=label_studio_access_token
    )
    logger.debug(f'Image stored in the local path: {image_local_path}')
    return image_local_path


def get_image_size(filepath):
    img = Image.open(filepath)
    img = ImageOps.exif_transpose(img)
    return img.size


class InMemoryLRUDictCache:
    def __init__(self, capacity=1):
        self.cache = OrderedDict()
        self.capacity = capacity

    def __contains__(self, item):
        return item in self.cache

    def get(self, key):
        if key in self.cache:
            # Move the accessed item to the end
            self.cache.move_to_end(key)
            return self.cache[key]
        return None

    def put(self, key, value):
        if key in self.cache:
            # Move the updated item to the end
            self.cache.move_to_end(key)
        elif len(self.cache) >= self.capacity:
            # Pop the first item if cache reached its capacity
            self.cache.popitem(last=False)

        self.cache[key] = value

    def __str__(self):
        return str(self.cache)


def match_labels(input: str, labels: List[str]) -> List[str]:
    # assuming classes are separated by newlines - we can customize that in the future
    # TODO: support other guardrails
    predicted_classes = input.splitlines()

    matched_labels = []
    for pred in predicted_classes:
        scores = list(map(lambda l: difflib.SequenceMatcher(None, pred, l).ratio(), labels))
        matched_labels.append(labels[scores.index(max(scores))])
    return matched_labels


if __name__ == "__main__":
    c = InMemoryLRUDictCache(2)
    c.put(1, 1)
    c.put(2,2)
    print(c.cache)


import setuptools
import lk_ml_package

print(lk_ml_package.package_name, lk_ml_package.__version__)

# Readme
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

# Module dependencies
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name=lk_ml_package.package_name,
    version=lk_ml_package.__version__,
    author='fengjiansong',
    author_email="fengjiansong@enn.cn",
    description='machine learning',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.enncloud.cn/ai_data_management/server/machine-learning.git',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'lk_ml_package=lk_ml_package.server:main'
        ],
    }
)
